from http.cookiejar import CookieJar
from typing import Any, Callable, List, Mapping, Sequence, Tuple, TypeVar, Union

from httpx import Cookies, Headers, QueryParams
from httpx._types import (
    AsyncByteStream,
    AuthTypes,
    CertTypes,
    ProxiesTypes,
    RequestContent,
    RequestData,
    RequestExtensions,
    RequestFiles,
    ResponseContent,
    ResponseExtensions,
    SyncByteStream,
    TimeoutTypes,
    URLTypes,
    VerifyTypes,
)
from typing_extensions import TypeAlias

__all__: Sequence[str] = (
    "AuthTypes",
    "CertTypes",
    "ProxiesTypes",
    "RequestContent",
    "RequestData",
    "RequestExtensions",
    "RequestFiles",
    "ResponseContent",
    "ResponseExtensions",
    "TimeoutTypes",
    "URLTypes",
    "VerifyTypes",
    "Primitive",
    "QueryTypes",
    "HeaderTypes",
    "CookieTypes",
    "PathTypes",
    "QueriesTypes",
    "HeadersTypes",
    "CookiesTypes",
    "PathsTypes",
    "MethodTypes",
    "JsonTypes",
    "StreamTypes",
    "EventHook",
    "EventHooks",
    "DefaultEncodingTypes",
)

StrOrBytes = TypeVar("StrOrBytes", str, bytes)

Primitive: TypeAlias = Union[
    str,
    int,
    float,
    bool,
    None,
]
QueryTypes: TypeAlias = Any
HeaderTypes: TypeAlias = Any
CookieTypes: TypeAlias = Any
PathTypes: TypeAlias = Union[Primitive, Sequence[Primitive]]
QueriesTypes: TypeAlias = Union[
    QueryParams,
    Mapping[str, Union[Primitive, Sequence[Primitive]]],
    List[Tuple[str, Primitive]],
    Tuple[Tuple[str, Primitive], ...],
    str,
    bytes,
]
HeadersTypes: TypeAlias = Union[
    Headers,
    Mapping[StrOrBytes, StrOrBytes],
    Sequence[Tuple[StrOrBytes, StrOrBytes]],
]
CookiesTypes: TypeAlias = Union[
    Cookies,
    CookieJar,
    Mapping[str, str],
    Sequence[Tuple[str, str]],
]
PathsTypes: TypeAlias = Union[
    Mapping[str, PathTypes],
    Sequence[Tuple[str, PathTypes]],
]

MethodTypes: TypeAlias = Union[str, bytes]
JsonTypes: TypeAlias = Any
StreamTypes: TypeAlias = Union[SyncByteStream, AsyncByteStream]
EventHook: TypeAlias = Callable[..., Any]
EventHooks: TypeAlias = Mapping[str, List[EventHook]]
DefaultEncodingTypes: TypeAlias = Union[str, Callable[[bytes], str]]



import inspect
from typing import Any, Callable, Dict, Tuple, Type

from .client import Client
from .models import ClientOptions
from .operation import Operation, get_operation, has_operation


class ServiceMeta(type):
    def __new__(
        cls: Type["ServiceMeta"], name: str, bases: Tuple[type], attrs: Dict[str, Any]
    ) -> type:
        def __init__(self) -> None:
            self._client = Client(client=self._opts.build())

            member_name: str
            member: Any
            for member_name, member in inspect.getmembers(self):
                if not has_operation(member):
                    continue

                bound_operation_func: Callable = self._client.bind(member)
                bound_operation_method: Callable = bound_operation_func.__get__(self)

                operation: Operation = get_operation(bound_operation_method)

                operation.func = bound_operation_method

                setattr(self, member_name, bound_operation_method)

        attrs["_opts"] = ClientOptions()
        attrs["__init__"] = __init__

        typ: type = super().__new__(cls, name, bases, attrs)

        return typ


class Service(metaclass=ServiceMeta):
    _opts: ClientOptions
    _client: Client

    def __repr__(self) -> str:
        return f"{type(self).__name__}()"



from dataclasses import dataclass, field
from typing import Optional, Protocol, Sequence

import mediate

from .enums import HeaderName
from .errors import ExpectedHeaderError, ExpectedStatusCodeError
from .models import Request, Response

__all__: Sequence[str] = (
    "Middleware",
    "RequestMiddleware",
    "ExpectedStatusCodeMiddleware",
    "raise_for_status",
)


class Middleware(mediate.Middleware[Request, Response]):
    pass


class RequestMiddleware(Protocol):
    def __call__(self, request: Request, /) -> Response:
        ...


@dataclass(init=False)
class ExpectedStatusCodeMiddleware:
    codes: Sequence[int] = field(default_factory=list)

    def __init__(self, *codes: int) -> None:
        self.codes = codes

    def __call__(self, call_next: RequestMiddleware, request: Request, /) -> Response:
        response: Response = call_next(request)

        if response.status_code not in self.codes:
            raise ExpectedStatusCodeError(
                f"Response contained an unexpected status code: {response.status_code}"
            )

        return response


@dataclass
class ExpectedHeaderMiddleware:
    name: str
    value: Optional[str] = None

    def __call__(self, call_next: RequestMiddleware, request: Request, /) -> Response:
        response: Response = call_next(request)

        header_value: str = response.headers[self.name]

        if self.name not in response.headers:
            raise ExpectedHeaderError(name=self.name)
        elif self.value is not None and header_value != self.value:
            raise ExpectedHeaderError(
                name=self.name, value=header_value, expected_value=self.value
            )

        return response


@dataclass
class ExpectedContentTypeMiddleware:
    content_type: str

    def __call__(self, call_next: RequestMiddleware, request: Request, /) -> Response:
        # NOTE: In the future this should parse the content type to support
        # lenient checking.
        # For example, 'application/json+foo' should likely be an acceptable
        # value when expecting a lenient form of 'application/json'
        return ExpectedHeaderMiddleware(HeaderName.CONTENT_TYPE, self.content_type)(
            call_next, request
        )


def raise_for_status(call_next: RequestMiddleware, request: Request, /) -> Response:
    response: Response = call_next(request)

    response.raise_for_status()

    return response



from http.cookiejar import CookieJar
from typing import Any, List, Mapping, MutableMapping, MutableSequence, Sequence, Tuple

from httpx import Cookies, Headers, QueryParams, Timeout
from httpx._utils import primitive_value_to_str

from .errors import ConversionError
from .types import (
    CookiesTypes,
    CookieTypes,
    HeadersTypes,
    HeaderTypes,
    PathsTypes,
    PathTypes,
    Primitive,
    QueriesTypes,
    QueryTypes,
    TimeoutTypes,
)
from .utils import is_primitive

__all__: Sequence[str] = (
    "convert_query_param",
    "convert_header",
    "convert_cookie",
    "convert_path_param",
    "convert_query_params",
    "convert_headers",
    "convert_cookies",
    "convert_path_params",
    "convert_timeout",
)


def convert_query_param(value: QueryTypes, /) -> str:
    if is_primitive(value):
        return primitive_value_to_str(value)
    else:
        return str(value)


def convert_header(value: HeaderTypes, /) -> str:
    if is_primitive(value):
        return primitive_value_to_str(value)
    else:
        return str(value)


def convert_cookie(value: CookieTypes, /) -> str:
    if is_primitive(value):
        return primitive_value_to_str(value)
    else:
        return str(value)


def convert_path_param(value: PathTypes, /) -> str:
    if isinstance(value, (str, int, float, bool)) or value is None:
        return primitive_value_to_str(value)
    elif isinstance(value, Sequence):
        segments: MutableSequence = []

        segment: Primitive
        for segment in value:
            converted_segment: str = primitive_value_to_str(segment)

            if converted_segment:
                segments.append(converted_segment)

        return "/".join(segments)
    else:
        raise ConversionError("path param", value)


def convert_query_params(value: QueriesTypes, /) -> QueryParams:
    if isinstance(value, QueryParams):
        return value
    elif isinstance(value, (Mapping, list, tuple)):
        return QueryParams(value)
    else:
        raise ConversionError("query params", value)


def convert_headers(value: HeadersTypes, /) -> Headers:
    if isinstance(value, Headers):
        return value
    elif isinstance(value, (Mapping, Sequence)):
        return Headers(dict(value))
    else:
        raise ConversionError("headers", value)


def convert_cookies(value: CookiesTypes, /) -> Cookies:
    if isinstance(value, Cookies):
        return value
    elif isinstance(value, CookieJar):
        return Cookies(value)
    elif isinstance(value, (Mapping, Sequence)):
        return Cookies(dict(value))
    else:
        raise ConversionError("cookies", value)


def convert_path_params(path_params: PathsTypes, /) -> MutableMapping[str, str]:
    if isinstance(path_params, Mapping):
        return {key: convert_path_param(value) for key, value in path_params.items()}
    else:
        return {key: convert_path_param(value) for key, value in path_params}


def convert_timeout(value: TimeoutTypes, /) -> Timeout:
    return Timeout(value)



from dataclasses import dataclass
from typing import Any, Callable, Optional, Protocol, Sequence, TypeVar

from typing_extensions import ParamSpec

from .client import NeoClient
from .enums import HttpMethod

__all__: Sequence[str] = (
    "request",
    "put",
    "get",
    "post",
    "head",
    "patch",
    "delete",
    "options",
)

PS = ParamSpec("PS")
RT = TypeVar("RT")


class OperationDecorator(Protocol):
    def __call__(
        self, endpoint: str, /, *, response: Optional[Callable[..., Any]] = None
    ) -> Callable[[Callable[PS, RT]], Callable[PS, RT]]:
        ...


@dataclass
class MethodOperationDecorator(OperationDecorator):
    method: str

    def __call__(
        self, endpoint: str, /, *, response: Optional[Callable[..., Any]] = None
    ) -> Callable[[Callable[PS, RT]], Callable[PS, RT]]:
        return NeoClient().request(self.method, endpoint, response=response)


def request(
    method: str, endpoint: str, /, *, response: Optional[Callable[..., Any]] = None
) -> Callable[[Callable[PS, RT]], Callable[PS, RT]]:
    return MethodOperationDecorator(method)(endpoint, response=response)


put: OperationDecorator = MethodOperationDecorator(HttpMethod.PUT)
get: OperationDecorator = MethodOperationDecorator(HttpMethod.GET)
post: OperationDecorator = MethodOperationDecorator(HttpMethod.POST)
head: OperationDecorator = MethodOperationDecorator(HttpMethod.HEAD)
patch: OperationDecorator = MethodOperationDecorator(HttpMethod.PATCH)
delete: OperationDecorator = MethodOperationDecorator(HttpMethod.DELETE)
options: OperationDecorator = MethodOperationDecorator(HttpMethod.OPTIONS)



import dataclasses
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Mapping,
    MutableMapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
)

import httpx
from httpx import URL, Cookies, Headers, QueryParams
from pydantic import BaseModel
from pydantic.fields import FieldInfo, ModelField

from . import api, utils
from .errors import PreparationError, ResolutionError
from .models import Request, Response
from .params import (
    BodyParameter,
    CookiesParameter,
    HeadersParameter,
    Parameter,
    QueriesParameter,
    QueryParameter,
    RequestParameter,
    ResponseParameter,
    URLParameter,
)
from .typing import Resolver
from .validation import ValidatedFunction

T = TypeVar("T")


def get_fields(func: Callable, /) -> Mapping[str, Tuple[Any, Parameter]]:
    class Config:
        allow_population_by_field_name: bool = True
        arbitrary_types_allowed: bool = True

    httpx_lookup: Mapping[Type[Any], Type[Parameter]] = {
        Request: RequestParameter,
        Response: ResponseParameter,
        httpx.Request: RequestParameter,
        httpx.Response: ResponseParameter,
        URL: URLParameter,
        QueryParams: QueriesParameter,
        Headers: HeadersParameter,
        Cookies: CookiesParameter,
    }

    fields: MutableMapping[str, Tuple[Any, Parameter]] = {}

    field_name: str
    model_field: ModelField
    for field_name, model_field in ValidatedFunction(
        func, config=Config
    ).model.__fields__.items():
        field_info: FieldInfo = model_field.field_info
        parameter: Parameter

        if not isinstance(field_info, Parameter):
            if model_field.annotation in httpx_lookup:
                parameter = httpx_lookup[model_field.annotation]()
            elif (
                isinstance(model_field.annotation, type)
                and issubclass(model_field.annotation, (BaseModel, dict))
                or dataclasses.is_dataclass(model_field.annotation)
            ):
                parameter = BodyParameter(
                    default=utils.get_default(field_info),
                )
            else:
                parameter = QueryParameter(
                    default=utils.get_default(field_info),
                )
        else:
            parameter = field_info

        # Create a clone of the parameter so that any mutations do not affect the original
        parameter_clone: Parameter = dataclasses.replace(parameter)

        parameter_clone.prepare(model_field)

        fields[field_name] = (model_field.annotation, parameter_clone)

    return fields


@dataclass
class DependencyResolver(Resolver[T]):
    dependency: Callable[..., T]

    def __call__(
        self,
        response: Response,
        /,
        *,
        cache: Optional[MutableMapping[Parameter, Any]] = None,
    ) -> T:
        if cache is None:
            cache = {}

        fields: Mapping[str, Tuple[Any, Parameter]] = get_fields(self.dependency)

        model_cls: Type[BaseModel] = api.create_model_cls(self.dependency, fields)

        arguments: MutableMapping[str, Any] = {}

        field_name: str
        parameter: Parameter
        for field_name, (_, parameter) in fields.items():
            resolution: Any

            if parameter in cache:
                resolution = cache[parameter]
            else:
                cache_parameter: bool = True

                if isinstance(parameter, DependencyParameter):
                    resolution = parameter.resolve(
                        response,
                        cache=cache,
                    )

                    cache_parameter = parameter.use_cache
                else:
                    resolution = parameter.resolve(response)

                if cache_parameter:
                    cache[parameter] = resolution

            # If there is no resolution (e.g. missing header/query param etc.)
            # and the parameter has a default, then we can omit the value from
            # the arguments.
            # This is done so that Pydantic will use the default value, rather
            # than complaining that None was used.
            if resolution is None and utils.has_default(parameter):
                continue

            arguments[field_name] = resolution

        model: BaseModel = model_cls(**arguments)

        validated_arguments: Mapping[str, Any] = model.dict()

        args: Tuple[Any, ...]
        kwargs: Mapping[str, Any]
        args, kwargs = utils.unpack_arguments(self.dependency, validated_arguments)

        return self.dependency(*args, **kwargs)


@dataclass(unsafe_hash=True)
class DependencyParameter(Parameter):
    dependency: Optional[Callable] = None
    use_cache: bool = True

    def resolve(
        self,
        response: Response,
        /,
        *,
        cache: Optional[MutableMapping[Parameter, Any]] = None,
    ) -> Any:
        if self.dependency is None:
            raise ResolutionError(
                f"Cannot resolve parameter {type(self)!r} without a dependency"
            )

        return DependencyResolver(self.dependency)(response, cache=cache)

    def prepare(self, field: ModelField, /) -> None:
        if self.dependency is not None:
            return

        # NOTE: The annotation will nearly always be callable (e.g. `int`)
        # This check needs to be changed to check for non primitive callables,
        # and more generally, nothing out of the standard library.
        if not callable(field.annotation):
            raise PreparationError(
                f"Failed to prepare parameter: {self!r}. Dependency has non-callable annotation"
            )

        self.dependency = field.annotation



import inspect
from functools import cached_property, wraps
from inspect import Parameter, Signature, signature
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from pydantic import Required
from pydantic.config import Extra
from pydantic.main import BaseModel, create_model
from pydantic.typing import get_all_type_hints
from pydantic.utils import to_camel
from typing_extensions import ParamSpec

__all__: Sequence[str] = (
    "create_func_model",
    "validate",
    "ValidatedFunction",
)

PS = ParamSpec("PS")
RT = TypeVar("RT")

ConfigType = Union[Type[Any], Mapping[str, Any]]


@overload
def validate(
    func: None = None, *, config: Optional[ConfigType] = None
) -> Callable[[Callable[PS, RT]], Callable[PS, RT]]:
    ...


@overload
def validate(func: Callable[PS, RT]) -> Callable[PS, RT]:
    ...


def validate(
    func: Optional[Callable[PS, RT]] = None, *, config: Optional[ConfigType] = None
) -> Union[Callable[PS, RT], Callable[[Callable[PS, RT]], Callable[PS, RT]]]:
    def decorate(func: Callable[PS, RT], /) -> Callable[PS, RT]:
        validated_func: ValidatedFunction[PS, RT] = ValidatedFunction(
            func, config=config
        )

        @wraps(func)
        def wrapper(*args: PS.args, **kwargs: PS.kwargs) -> RT:
            return validated_func.call(*args, **kwargs)

        setattr(wrapper, "validator", validated_func)

        return wrapper

    if func:
        return decorate(func)
    else:
        return decorate


def create_func_model(
    function: Callable,
    fields: Mapping[str, Any],
    *,
    config: Optional[ConfigType] = None,
) -> Type[BaseModel]:
    configurations: Dict[str, Any] = {}

    if isinstance(config, dict):
        configurations.update(config)
    elif isinstance(config, type):
        member_name: str
        member: Any
        for member_name, member in inspect.getmembers(config):
            if member_name.startswith("_"):
                continue

            configurations[member_name] = member

    configurations.setdefault("extra", Extra.forbid)

    Config: Type[Any] = type("Config", (), configurations)

    ValidatedFunctionBaseModel: Type[BaseModel] = type(
        "ValidatedFunctionBaseModel", (BaseModel,), {"Config": Config}
    )

    return create_model(
        to_camel(function.__name__),
        __base__=ValidatedFunctionBaseModel,
        **fields,
    )


class ValidatedFunction(Generic[PS, RT]):
    function: Callable[PS, RT]
    model: Type[BaseModel]

    def __init__(
        self, function: Callable[PS, RT], /, *, config: Optional[ConfigType] = None
    ):
        self.function = function

        parameters: Mapping[str, Parameter] = self.signature.parameters
        type_hints: Mapping[str, Any] = get_all_type_hints(function)
        fields: MutableMapping[str, Tuple[Any, Any]] = {}

        parameter_name: str
        parameter: Parameter
        for parameter_name, parameter in parameters.items():
            annotation: Any = (
                Any
                if parameter.annotation is Parameter.empty
                else type_hints[parameter_name]
            )

            default: Any = (
                Required if parameter.default is Parameter.empty else parameter.default
            )

            if parameter.kind is Parameter.VAR_POSITIONAL:
                fields[parameter_name] = (Tuple[annotation, ...], None)
            elif parameter.kind is Parameter.VAR_KEYWORD:
                fields[parameter_name] = (Mapping[str, annotation], None)
            else:
                fields[parameter_name] = (annotation, default)

        self.model = create_func_model(function, fields, config=config)

    def __repr__(self) -> str:
        return f"<{type(self).__name__}(function={self.function!r})>"

    @cached_property
    def signature(self) -> Signature:
        return signature(self.function)

    def bind_arguments(
        self, args: Tuple[Any, ...], kwargs: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        return self.signature.bind(*args, **kwargs).arguments

    def prepare_arguments(
        self, arguments: Mapping[str, Any], /
    ) -> Tuple[Tuple[Any, ...], Mapping[str, Any]]:
        args: MutableSequence[Any] = []
        kwargs: MutableMapping[str, Any] = {}

        parameter: Parameter
        for parameter in self.signature.parameters.values():
            argument: Any = arguments[parameter.name]

            if parameter.kind in (
                Parameter.POSITIONAL_ONLY,
                Parameter.POSITIONAL_OR_KEYWORD,
            ):
                args.append(argument)
            elif parameter.kind is Parameter.KEYWORD_ONLY:
                kwargs[parameter.name] = argument
            elif parameter.kind is Parameter.VAR_POSITIONAL:
                args.extend(argument)
            elif parameter.kind is Parameter.VAR_KEYWORD:
                kwargs.update(argument)

        return (tuple(args), kwargs)

    def validate_arguments(
        self, args: Tuple[Any, ...], kwargs: Mapping[str, Any]
    ) -> Tuple[Tuple[Any, ...], Mapping[str, Any]]:
        bound_arguments: Mapping[str, Any] = self.bind_arguments(args, kwargs)
        model: BaseModel = self.model(**bound_arguments)
        arguments: Mapping[str, Any] = model.dict()

        return self.prepare_arguments(arguments)

    def call(self, *args: PS.args, **kwargs: PS.kwargs) -> RT:
        validated_args: Tuple[Any, ...]
        validated_kwargs: Mapping[str, Any]
        validated_args, validated_kwargs = self.validate_arguments(args, kwargs)

        return self.function(*validated_args, **validated_kwargs)



from dataclasses import dataclass
from typing import Any, Optional, Sequence

__all__: Sequence[str] = (
    "DuplicateParameters",
    "IncompatiblePathParameters",
    "PreparationError",
    "CompositionError",
    "ResolutionError",
    "ConversionError",
    "NotAnOperationError",
    "ExpectedStatusCodeError",
    "ExpectedHeaderError",
    "MissingStateError",
)


class DuplicateParameters(Exception):
    pass


class IncompatiblePathParameters(Exception):
    pass


class PreparationError(Exception):
    pass


class CompositionError(Exception):
    pass


class ResolutionError(Exception):
    pass


@dataclass
class ConversionError(Exception):
    subject: str
    value: Any

    def __str__(self) -> str:
        return f"Cannot convert {self.subject} from type {type(self.value)}"


class NotAnOperationError(Exception):
    pass


class ExpectedStatusCodeError(Exception):
    pass


@dataclass
class ExpectedHeaderError(Exception):
    name: str
    value: Optional[str] = None
    expected_value: Optional[str] = None

    def __str__(self) -> str:
        if self.value is None:
            return f"Response missing expected header {str(self.name)!r}"
        else:
            return f"Response header {str(self.name)!r} has incorrect value. Expected {self.expected_value!r}, got {self.value!r}"


@dataclass
class MissingStateError(Exception):
    key: str

    def __str__(self) -> str:
        return f"State entry missing for key {self.key!r}"



from typing import Any, Protocol, Sequence, TypeVar

from .models import ClientOptions, PreRequest, Request, Response

__all__: Sequence[str] = (
    "CallNext",
    "ClientConsumer",
    "Composer",
    "RequestConsumer",
    "Resolver",
    "Supplier",
)

T_contra = TypeVar("T_contra", contravariant=True)
T_co = TypeVar("T_co", covariant=True)
R_co = TypeVar("R_co", covariant=True)


class CallNext(Protocol):
    def __call__(self, request: Request, /) -> Response:
        ...


class Supplier(Protocol[T_co]):
    def __call__(self) -> T_co:
        ...


class Resolver(Protocol[T_co]):
    def __call__(self, response: Response, /) -> T_co:
        ...


class Composer(Protocol):
    def compose(self, request: PreRequest, argument: Any, /) -> None:
        ...


class RequestConsumer(Protocol):
    def __call__(self, request: PreRequest, /) -> None:
        ...


class ClientConsumer(Protocol):
    def __call__(self, client: ClientOptions, /) -> None:
        ...



from dataclasses import dataclass
from typing import Callable, Protocol, Sequence, Type, TypeVar

from mediate.protocols import MiddlewareCallable

from .consumers import (
    BaseURLConsumer,
    Consumer,
    ContentConsumer,
    CookieConsumer,
    CookiesConsumer,
    DataConsumer,
    FilesConsumer,
    HeaderConsumer,
    HeadersConsumer,
    JsonConsumer,
    MountConsumer,
    PathConsumer,
    PathsConsumer,
    QueriesConsumer,
    QueryConsumer,
    TimeoutConsumer,
)
from .enums import HeaderName
from .errors import CompositionError
from .models import ClientOptions, PreRequest, Request, Response
from .operation import OperationSpecification, get_operation
from .service import Service
from .types import (
    CookiesTypes,
    CookieTypes,
    HeadersTypes,
    HeaderTypes,
    JsonTypes,
    PathsTypes,
    PathTypes,
    QueriesTypes,
    QueryTypes,
    RequestContent,
    RequestData,
    RequestFiles,
    TimeoutTypes,
)

__all__: Sequence[str] = (
    "query",
    "header",
    "cookie",
    "path",
    "query_params",
    "headers",
    "cookies",
    "path_params",
    "content",
    "data",
    "files",
    "json",
    "timeout",
)

T = TypeVar("T", Callable, Type[Service])


class Decorator(Protocol):
    def __call__(self, target: T, /) -> T:
        ...


@dataclass
class CompositionDecorator(Decorator):
    consumer: Consumer

    def __call__(self, target: T, /) -> T:
        if isinstance(target, type):
            if not issubclass(target, Service):
                raise CompositionError(f"Target class is not a subclass of {Service}")

            client: ClientOptions = target._opts

            self.consumer.consume_client(client)
        elif callable(target):
            request: PreRequest = get_operation(target).specification.request

            self.consumer.consume_request(request)
        else:
            raise CompositionError(f"Target of unsupported type {type(target)}")

        return target


def query(key: str, value: QueryTypes) -> Decorator:
    return CompositionDecorator(QueryConsumer(key, value))


def header(key: str, value: HeaderTypes) -> Decorator:
    return CompositionDecorator(HeaderConsumer(key, value))


def cookie(key: str, value: CookieTypes) -> Decorator:
    return CompositionDecorator(CookieConsumer(key, value))


def path(key: str, value: PathTypes) -> Decorator:
    return CompositionDecorator(PathConsumer(key, value))


def query_params(params: QueriesTypes, /) -> Decorator:
    return CompositionDecorator(QueriesConsumer(params))


def headers(headers: HeadersTypes, /) -> Decorator:
    return CompositionDecorator(HeadersConsumer(headers))


def cookies(cookies: CookiesTypes, /) -> Decorator:
    return CompositionDecorator(CookiesConsumer(cookies))


def path_params(path_params: PathsTypes, /) -> Decorator:
    return CompositionDecorator(PathsConsumer(path_params))


def content(content: RequestContent, /) -> Decorator:
    return CompositionDecorator(ContentConsumer(content))


def data(data: RequestData, /) -> Decorator:
    return CompositionDecorator(DataConsumer(data))


def files(files: RequestFiles, /) -> Decorator:
    return CompositionDecorator(FilesConsumer(files))


def json(json: JsonTypes, /) -> Decorator:
    return CompositionDecorator(JsonConsumer(json))


def timeout(timeout: TimeoutTypes, /) -> Decorator:
    return CompositionDecorator(TimeoutConsumer(timeout))


def mount(path: str, /) -> Decorator:
    return CompositionDecorator(MountConsumer(path))


def base_url(base_url: str, /) -> Decorator:
    return CompositionDecorator(BaseURLConsumer(base_url))


def middleware(*middleware: MiddlewareCallable[Request, Response]) -> Decorator:
    def decorate(target: T, /) -> T:
        if isinstance(target, type) and issubclass(target, Service):
            raise CompositionError(
                "Middleware decorator currently unsupported for service classes"
            )

        specification: OperationSpecification = get_operation(target).specification

        middleware_callable: MiddlewareCallable[Request, Response]
        for middleware_callable in middleware:
            specification.middleware.add(middleware_callable)

        return target

    return decorate


def user_agent(user_agent: str, /) -> Decorator:
    return header(HeaderName.USER_AGENT, user_agent)


def accept(*content_types: str) -> Decorator:
    return header(HeaderName.ACCEPT, ",".join(content_types))


def referer(referer: str, /) -> Decorator:
    return header(HeaderName.REFERER, referer)



from typing import Any, Callable

from .dependencies import DependencyParameter
from .models import Response


def resolve(func: Callable, response: Response) -> Any:
    dependency: DependencyParameter = DependencyParameter(dependency=func)

    return dependency.resolve(response)



from dataclasses import dataclass
from typing import Any, Optional, Sequence, TypeVar

from httpx import Cookies, Headers, QueryParams

from .models import Response
from .typing import Resolver

__all__: Sequence[str] = (
    "BodyResolver",
    "CookieResolver",
    "CookiesResolver",
    "HeaderResolver",
    "HeadersResolver",
    "QueryResolver",
    "QueriesResolver",
    "StateResolver",
)

T = TypeVar("T")


@dataclass
class QueryResolver(Resolver[Optional[str]]):
    name: str

    def __call__(self, response: Response, /) -> Optional[str]:
        return response.request.url.params.get(self.name)


@dataclass
class HeaderResolver(Resolver[Optional[str]]):
    name: str

    def __call__(self, response: Response, /) -> Optional[str]:
        return response.headers.get(self.name)


@dataclass
class CookieResolver(Resolver[Optional[str]]):
    name: str

    def __call__(self, response: Response, /) -> Optional[str]:
        return response.cookies.get(self.name)


class QueriesResolver(Resolver[QueryParams]):
    @staticmethod
    def __call__(response: Response, /) -> QueryParams:
        return response.request.url.params


class HeadersResolver(Resolver[Headers]):
    @staticmethod
    def __call__(response: Response, /) -> Headers:
        return response.headers


class CookiesResolver(Resolver[Cookies]):
    @staticmethod
    def __call__(response: Response, /) -> Cookies:
        return response.cookies


class BodyResolver(Resolver[Any]):
    @staticmethod
    def __call__(response: Response, /) -> Any:
        return response.json()


@dataclass
class StateResolver(Resolver[Any]):
    key: str

    def __call__(self, response: Response, /) -> Any:
        return response.state.get(self.key)



from dataclasses import dataclass
from typing import Any, Mapping, Sequence, Union

from httpx import URL, Cookies, Headers, QueryParams, Timeout

from . import converters
from .errors import CompositionError
from .models import ClientOptions, PreRequest
from .types import (
    CookieTypes,
    HeaderTypes,
    JsonTypes,
    PathsTypes,
    PathTypes,
    QueriesTypes,
    QueryTypes,
    RequestContent,
    RequestData,
    RequestFiles,
    TimeoutTypes,
)

__all__: Sequence[str] = (
    "QueryConsumer",
    "HeaderConsumer",
    "CookieConsumer",
    "PathConsumer",
    "QueriesConsumer",
    "HeadersConsumer",
    "CookiesConsumer",
    "PathsConsumer",
    "ContentConsumer",
    "DataConsumer",
    "FilesConsumer",
    "JsonConsumer",
    "TimeoutConsumer",
    "MountConsumer",
    "BaseURLConsumer",
)


class Consumer:
    def consume(self, target: Union[PreRequest, ClientOptions], /) -> None:
        if isinstance(target, PreRequest):
            self.consume_request(target)
        elif isinstance(target, ClientOptions):
            self.consume_client(target)
        else:
            raise CompositionError(
                f"Consumer {type(self).__name__!r} does not support consumption of type {type(target)}"
            )

    def consume_request(self, _: PreRequest, /) -> None:
        raise CompositionError(
            f"Consumer {type(self).__name__!r} does not support consumption of type {PreRequest}"
        )

    def consume_client(self, _: ClientOptions, /) -> None:
        raise CompositionError(
            f"Consumer {type(self).__name__!r} does not support consumption of type {ClientOptions}"
        )


@dataclass(init=False)
class QueryConsumer(Consumer):
    key: str
    value: str

    def __init__(self, key: str, value: QueryTypes) -> None:
        self.key = key
        self.value = converters.convert_query_param(value)

    def consume_request(self, request: PreRequest, /) -> None:
        request.params = request.params.set(self.key, self.value)

    def consume_client(self, client: ClientOptions, /) -> None:
        client.params = client.params.set(self.key, self.value)


@dataclass(init=False)
class HeaderConsumer(Consumer):
    key: str
    value: str

    def __init__(self, key: str, value: HeaderTypes) -> None:
        self.key = key
        self.value = converters.convert_header(value)

    def consume_request(self, request: PreRequest, /) -> None:
        request.headers[self.key] = self.value

    def consume_client(self, client: ClientOptions, /) -> None:
        client.headers[self.key] = self.value


@dataclass(init=False)
class CookieConsumer(Consumer):
    key: str
    value: str

    def __init__(self, key: str, value: CookieTypes) -> None:
        self.key = key
        self.value = converters.convert_cookie(value)

    def consume_request(self, request: PreRequest, /) -> None:
        request.cookies[self.key] = self.value

    def consume_client(self, client: ClientOptions, /) -> None:
        client.cookies[self.key] = self.value


@dataclass(init=False)
class PathConsumer(Consumer):
    key: str
    value: str

    def __init__(self, key: str, value: PathTypes) -> None:
        self.key = key
        self.value = converters.convert_path_param(value)

    def consume_request(self, request: PreRequest, /) -> None:
        request.path_params[self.key] = self.value


@dataclass(init=False)
class QueriesConsumer(Consumer):
    params: QueryParams

    def __init__(self, params: QueriesTypes, /) -> None:
        self.params = converters.convert_query_params(params)

    def consume_request(self, request: PreRequest, /) -> None:
        request.params = request.params.merge(self.params)

    def consume_client(self, client: ClientOptions, /) -> None:
        client.params = client.params.merge(self.params)


@dataclass(init=False)
class HeadersConsumer(Consumer):
    headers: Headers

    def __init__(self, headers: HeaderTypes, /) -> None:
        self.headers = converters.convert_headers(headers)

    def consume_request(self, request: PreRequest, /) -> None:
        request.headers.update(self.headers)

    def consume_client(self, client: ClientOptions, /) -> None:
        client.headers.update(self.headers)


@dataclass(init=False)
class CookiesConsumer(Consumer):
    cookies: Cookies

    def __init__(self, cookies: CookieTypes, /) -> None:
        self.cookies = converters.convert_cookies(cookies)

    def consume_request(self, request: PreRequest, /) -> None:
        request.cookies.update(self.cookies)

    def consume_client(self, client: ClientOptions, /) -> None:
        client.cookies.update(self.cookies)


@dataclass(init=False)
class PathsConsumer(Consumer):
    path_params: Mapping[str, str]

    def __init__(self, path_params: PathsTypes, /) -> None:
        self.path_params = converters.convert_path_params(path_params)

    def consume_request(self, request: PreRequest, /) -> None:
        request.path_params.update(self.path_params)


@dataclass
class ContentConsumer(Consumer):
    content: RequestContent

    def consume_request(self, request: PreRequest, /) -> None:
        request.content = self.content


@dataclass
class DataConsumer(Consumer):
    data: RequestData

    def consume_request(self, request: PreRequest, /) -> None:
        request.data = self.data


@dataclass
class FilesConsumer(Consumer):
    files: RequestFiles

    def consume_request(self, request: PreRequest, /) -> None:
        request.files = self.files


@dataclass
class JsonConsumer(Consumer):
    json: JsonTypes

    def consume_request(self, request: PreRequest, /) -> None:
        request.json = self.json


@dataclass(init=False)
class TimeoutConsumer(Consumer):
    timeout: Timeout

    def __init__(self, timeout: TimeoutTypes, /) -> None:
        self.timeout = converters.convert_timeout(timeout)

    def consume_request(self, request: PreRequest, /) -> None:
        request.timeout = self.timeout

    def consume_client(self, client: ClientOptions, /) -> None:
        client.timeout = self.timeout


@dataclass
class StateConsumer(Consumer):
    key: str
    value: Any

    def consume_request(self, request: PreRequest, /) -> None:
        request.state[self.key] = self.value


@dataclass
class MountConsumer(Consumer):
    path: str

    def consume_request(self, request: PreRequest, /) -> None:
        request.url = request.url.copy_with(path=self.path + request.url.path)


@dataclass
class BaseURLConsumer(Consumer):
    base_url: str

    def consume_client(self, client: ClientOptions, /) -> None:
        client.base_url = URL(self.base_url)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['neoclient']

package_data = \
{'': ['*']}

install_requires = \
['httpx>=0.23.3,<0.24.0',
 'mediate>=0.1.7,<0.2.0',
 'pydantic>=1.10.0,<2.0.0',
 'typing-extensions>=4.3.0,<5.0.0']

setup_kwargs = {
    'name': 'neoclient',
    'version': '0.1.19',
    'description': 'Fast API Clients for Python',
    'long_description': '# neoclient\n🚀 Fast API Clients for Python\n\n## Installation\n```console\npip install neoclient\n```\n\n## Introduction\nThe simplest `neoclient` file might look like this:\n```python\nfrom neoclient import get\n\n@get("https://httpbin.org/ip")\ndef ip():\n    ...\n```\n```python\n>>> ip()\n{\'origin\': \'1.2.3.4\'}\n```\n\nHowever, it\'s almost always better to create a `NeoClient` instance:\n```python\nfrom neoclient import NeoClient\n\nclient = NeoClient("https://httpbin.org/")\n\n@client.get("/ip")\ndef ip():\n    ...\n```\n```python\n>>> ip()\n{\'origin\': \'1.2.3.4\'}\n```',
    'author': 'Tom Bulled',
    'author_email': '26026015+tombulled@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/tombulled/neoclient',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import functools
import inspect
from dataclasses import dataclass, field
from json import JSONDecodeError
from typing import Any, Callable, Generic, Optional, Sequence, TypeVar

import httpx
import pydantic
from httpx import Client
from pydantic import BaseModel
from typing_extensions import ParamSpec

from .composition import compose
from .errors import NotAnOperationError
from .middleware import Middleware
from .models import PreRequest, Request, Response
from .resolution import resolve

__all__: Sequence[str] = (
    "set_operation",
    "has_operation",
    "get_operation",
    "Operation",
)

PS = ParamSpec("PS")
RT = TypeVar("RT", covariant=True)

ATTRIBUTE_OPERATION: str = "operation"


def set_operation(func: Callable, operation: "Operation", /) -> None:
    setattr(func, ATTRIBUTE_OPERATION, operation)


def has_operation(func: Callable, /) -> bool:
    return hasattr(func, ATTRIBUTE_OPERATION)


def get_operation(func: Callable, /) -> "Operation":
    operation: Optional[Operation] = getattr(func, ATTRIBUTE_OPERATION, None)

    if operation is None:
        raise NotAnOperationError(f"{func!r} is not an operation")

    return operation


@dataclass
class OperationSpecification:
    request: PreRequest
    response: Optional[Callable[..., Any]] = None
    middleware: Middleware = field(default_factory=Middleware)


@dataclass
class Operation(Generic[PS, RT]):
    func: Callable[PS, RT]
    specification: OperationSpecification
    client: Optional[Client]
    middleware: Middleware = field(default_factory=Middleware)

    def __call__(self, *args: PS.args, **kwargs: PS.kwargs) -> Any:
        pre_request: PreRequest = self.specification.request.merge(
            PreRequest(
                method=self.specification.request.method,
                url=self.specification.request.url,
            )
        )

        compose(self.func, pre_request, args, kwargs)

        request: Request = pre_request.build_request(self.client)

        return_annotation: Any = inspect.signature(self.func).return_annotation

        if return_annotation is PreRequest:
            return pre_request
        if return_annotation is Request:
            return request

        client: Client = self.client if self.client is not None else Client()

        middleware: Middleware = Middleware()

        middleware.record.extend(self.specification.middleware.record)
        middleware.record.extend(self.middleware.record)

        @middleware.compose
        def send_request(request: Request, /) -> Response:
            httpx_response: httpx.Response = client.send(request)

            return Response.from_httpx_response(httpx_response)

        response: Response = send_request(request)

        if self.specification.response is not None:
            return resolve(self.specification.response, response)

        if return_annotation is inspect.Parameter.empty:
            try:
                return response.json()
            except JSONDecodeError:
                return response.text
        if return_annotation is None:
            return None
        if return_annotation is Response:
            return response
        if isinstance(return_annotation, type) and issubclass(
            return_annotation, BaseModel
        ):
            return return_annotation.parse_obj(response.json())

        return pydantic.parse_raw_as(return_annotation, response.text)

    @property
    def wrapper(self) -> Callable[PS, RT]:
        @functools.wraps(self.func)
        def wrapper(*args: PS.args, **kwargs: PS.kwargs) -> RT:
            if inspect.ismethod(self.func):
                # Read off `self` or `cls`
                _, *args = args  # type: ignore

            return self(*args, **kwargs)

        set_operation(wrapper, self)

        return wrapper



import dataclasses
from dataclasses import dataclass
from typing import Callable, Dict, List, Optional, Sequence, TypeVar

import httpx
from httpx import URL, Cookies, Headers, QueryParams, Timeout
from httpx._auth import Auth
from mediate.protocols import MiddlewareCallable
from roster import Record
from typing_extensions import ParamSpec

from . import __version__, converters
from .composition import get_fields, validate_fields
from .defaults import (
    DEFAULT_AUTH,
    DEFAULT_BASE_URL,
    DEFAULT_COOKIES,
    DEFAULT_ENCODING,
    DEFAULT_EVENT_HOOKS,
    DEFAULT_FOLLOW_REDIRECTS,
    DEFAULT_HEADERS,
    DEFAULT_MAX_REDIRECTS,
    DEFAULT_PARAMS,
    DEFAULT_TIMEOUT,
    DEFAULT_TRUST_ENV,
)
from .enums import HeaderName, HttpMethod
from .middleware import Middleware
from .models import PreRequest, Request, Response
from .operation import Operation, OperationSpecification, get_operation
from .types import (
    AuthTypes,
    CookiesTypes,
    DefaultEncodingTypes,
    EventHooks,
    HeadersTypes,
    QueriesTypes,
    TimeoutTypes,
    URLTypes,
)

__all__: Sequence[str] = (
    "Session",
    "Client",
    "NeoClient",
)


T = TypeVar("T")

PS = ParamSpec("PS")
RT = TypeVar("RT")

USER_AGENT: str = f"neoclient/{__version__}"


class BaseService:
    def __repr__(self) -> str:
        return f"{type(self).__name__}()"


@dataclass(init=False)
class Session(httpx.Client):
    auth: Optional[Auth]
    params: QueryParams
    headers: Headers
    cookies: Cookies
    timeout: Timeout
    follow_redirects: bool
    max_redirects: int
    event_hooks: Dict[str, List[Callable]]
    base_url: URL
    trust_env: bool

    def __init__(
        self,
        base_url: URLTypes = DEFAULT_BASE_URL,
        *,
        auth: Optional[AuthTypes] = DEFAULT_AUTH,
        params: Optional[QueriesTypes] = DEFAULT_PARAMS,
        headers: Optional[HeadersTypes] = DEFAULT_HEADERS,
        cookies: Optional[CookiesTypes] = DEFAULT_COOKIES,
        timeout: TimeoutTypes = DEFAULT_TIMEOUT,
        follow_redirects: bool = DEFAULT_FOLLOW_REDIRECTS,
        max_redirects: int = DEFAULT_MAX_REDIRECTS,
        event_hooks: Optional[EventHooks] = DEFAULT_EVENT_HOOKS,
        trust_env: bool = DEFAULT_TRUST_ENV,
        default_encoding: DefaultEncodingTypes = DEFAULT_ENCODING,
    ) -> None:
        params = (
            converters.convert_query_params(params)
            if params is not None
            else QueryParams()
        )
        headers = (
            converters.convert_headers(headers) if headers is not None else Headers()
        )
        cookies = (
            converters.convert_cookies(cookies) if cookies is not None else Cookies()
        )
        timeout = (
            converters.convert_timeout(timeout) if timeout is not None else Timeout()
        )
        base_url = URL(base_url)

        headers.setdefault(HeaderName.USER_AGENT, USER_AGENT)

        super().__init__(
            auth=auth,
            params=params,
            headers=headers,
            cookies=cookies,
            timeout=timeout,
            follow_redirects=follow_redirects,
            max_redirects=max_redirects,
            event_hooks=event_hooks,
            base_url=base_url,
            trust_env=trust_env,
            default_encoding=default_encoding,
        )


@dataclass(init=False)
class Client:
    client: Optional[httpx.Client]
    middleware: Middleware

    def __init__(
        self,
        client: Optional[httpx.Client] = None,
        middleware: Optional[Middleware] = None,
    ) -> None:
        self.client = client
        self.middleware = middleware if middleware is not None else Middleware()

    def bind(self, func: Callable[PS, RT], /) -> Callable[PS, RT]:
        operation: Operation = get_operation(func)

        bound_operation: Operation[PS, RT] = dataclasses.replace(
            operation, client=self.client
        )

        return bound_operation.wrapper

    def request(
        self,
        method: str,
        endpoint: str,
        /,
        *,
        response: Optional[Callable] = None,
    ) -> Callable[[Callable[PS, RT]], Callable[PS, RT]]:
        specification: OperationSpecification = OperationSpecification(
            request=PreRequest(
                method=method,
                url=endpoint,
            ),
            response=response,
        )

        def decorator(func: Callable[PS, RT], /) -> Callable[PS, RT]:
            operation: Operation[PS, RT] = Operation(
                func=func,
                specification=specification,
                client=self.client,
                middleware=self.middleware,
            )

            # Validate operation function parameters are acceptable
            validate_fields(get_fields(specification.request, func))

            return operation.wrapper

        return decorator

    def put(
        self, endpoint: str, /, *, response: Optional[Callable] = None
    ) -> Callable[[Callable[PS, RT]], Callable[PS, RT]]:
        return self.request(HttpMethod.PUT.name, endpoint, response=response)

    def get(
        self, endpoint: str, /, *, response: Optional[Callable] = None
    ) -> Callable[[Callable[PS, RT]], Callable[PS, RT]]:
        return self.request(HttpMethod.GET.name, endpoint, response=response)

    def post(
        self, endpoint: str, /, *, response: Optional[Callable] = None
    ) -> Callable[[Callable[PS, RT]], Callable[PS, RT]]:
        return self.request(HttpMethod.POST.name, endpoint, response=response)

    def head(
        self, endpoint: str, /, *, response: Optional[Callable] = None
    ) -> Callable[[Callable[PS, RT]], Callable[PS, RT]]:
        return self.request(HttpMethod.HEAD.name, endpoint, response=response)

    def patch(
        self, endpoint: str, /, *, response: Optional[Callable] = None
    ) -> Callable[[Callable[PS, RT]], Callable[PS, RT]]:
        return self.request(HttpMethod.PATCH.name, endpoint, response=response)

    def delete(
        self, endpoint: str, /, *, response: Optional[Callable] = None
    ) -> Callable[[Callable[PS, RT]], Callable[PS, RT]]:
        return self.request(HttpMethod.DELETE.name, endpoint, response=response)

    def options(
        self, endpoint: str, /, *, response: Optional[Callable] = None
    ) -> Callable[[Callable[PS, RT]], Callable[PS, RT]]:
        return self.request(HttpMethod.OPTIONS.name, endpoint, response=response)


class NeoClient(Client):
    def __init__(
        self,
        base_url: URLTypes = DEFAULT_BASE_URL,
        *,
        auth: Optional[AuthTypes] = DEFAULT_AUTH,
        params: Optional[QueriesTypes] = DEFAULT_PARAMS,
        headers: Optional[HeadersTypes] = DEFAULT_HEADERS,
        cookies: Optional[CookiesTypes] = DEFAULT_COOKIES,
        timeout: TimeoutTypes = DEFAULT_TIMEOUT,
        follow_redirects: bool = DEFAULT_FOLLOW_REDIRECTS,
        max_redirects: int = DEFAULT_MAX_REDIRECTS,
        event_hooks: Optional[EventHooks] = DEFAULT_EVENT_HOOKS,
        trust_env: bool = DEFAULT_TRUST_ENV,
        default_encoding: DefaultEncodingTypes = DEFAULT_ENCODING,
        middleware: Optional[Sequence[MiddlewareCallable[Request, Response]]] = None,
    ) -> None:
        super().__init__(
            client=Session(
                auth=auth,
                params=params,
                headers=headers,
                cookies=cookies,
                timeout=timeout,
                follow_redirects=follow_redirects,
                max_redirects=max_redirects,
                event_hooks=event_hooks,
                base_url=base_url,
                trust_env=trust_env,
                default_encoding=default_encoding,
            ),
            middleware=(
                Middleware(record=Record(middleware))
                if middleware is not None
                else Middleware()
            ),
        )
